# """Test Classification Manager Module."""
# import pytest
# from geniepy.errors import ClassifierError
# import geniepy.datamgmt.daos as daos
# import geniepy.datamgmt.repositories as dr
# from geniepy.datamgmt.tables import PUBMED_PROPTY, CTD_PROPTY, CLSFR_PROPTY
# from geniepy.datamgmt import DaoManager
# from geniepy.datamgmt.parsers import ClassifierParser
# from tests.resources.mock import MOCK_CLSFRMGR, TEST_CHUNKSIZE
# import tests.resources.mock as mock


# class TestClassMgr:
#     """PyTest Class to test Classification manager."""

#     # Create and configure mock ctd dao
#     ctd_dao = daos.CtdDao(dr.SqlRepository("sqlite://", CTD_PROPTY))
#     # pylint: disable=protected-access
#     ctd_dao._parser.scraper = mock.MockCtdScraper()

#     # Create and configure mock pubmed dao
#     pubmed_dao = daos.PubMedDao(dr.SqlRepository("sqlite://", PUBMED_PROPTY))
#     # pylint: disable=protected-access
#     pubmed_dao._parser.scraper = mock.MockPubMedScraper()

#     # Create and configure mock pubmed dao
#     classifier_dao = daos.ClassifierDao(dr.SqlRepository("sqlite://", CLSFR_PROPTY))
#     # pylint: disable=protected-access

#     # Construct mock dao manager for testing
#     dao_mgr = DaoManager(
#         ctd_dao=ctd_dao, pubmed_dao=pubmed_dao, classifier_dao=classifier_dao
#     )

#     def test_constructor(self):
#         """Test obj construction."""
#         assert MOCK_CLSFRMGR is not None

#     def test_predict_records(self):
#         """
#         Test prediction of records.

#         Records are fed into the classifier to be predicted and classification manager
#         returns a dataframe containing the corresponding predictions.
#         """
#         # Generate records to be fed into classifiers
#         self.dao_mgr.download(TEST_CHUNKSIZE)
#         gen_df = self.dao_mgr.gen_records(TEST_CHUNKSIZE)
#         raw_df = next(gen_df)
#         predicted_df = MOCK_CLSFRMGR.predict(raw_df)
#         # Make sure predicted all rows
#         expected_rows = raw_df.shape[0]
#         actual_rows = predicted_df.shape[0]
#         assert actual_rows == expected_rows
#         # Make sure predicted df is valid (should return no errors)
#         assert not ClassifierParser.validate(predicted_df)
#         # Make sure one prediction per classifier
#         cols = predicted_df.columns
#         # Make sure has a digest column
#         assert "digest" in cols
#         # Make sure has one prediction column per classifier
#         for classifier in MOCK_CLSFRMGR._classifiers:
#             assert classifier.name in cols
#         # TODO validate classifier predicted dataframe

#     def test_predict_invalid_records(self):
#         """Test attempting to predict with invalid records."""
#         with pytest.raises(ClassifierError):
#             MOCK_CLSFRMGR.predict(None)

# def test_predict_invalid_records(self):
#     """Test attempting to predict with invalid records."""
#     with pytest.raises(ClassifierError):
#         MOCK_CLSFRMGR.predict(None)



"""Geniepy Errors Module."""


class GeniePyError(Exception):
    """General GeniePy error, all other geniepy errors inherits from GeniePyError."""

    def __init__(self, msg=""):
        """Add message geniepy exceptions."""
        self._msg = msg
        super().__init__()

    @property
    def message(self):
        """Return exception message."""
        return self._msg


class SchemaError(GeniePyError):
    """Invalid database schema."""


class ConnectionError(GeniePyError):
    """Unable to connect."""


class DaoError(GeniePyError):
    """Data Access Object Error."""


class ParserError(GeniePyError):
    """Unable to parse data."""


class ClassifierError(GeniePyError):
    """Classifier unable to execute command."""


class ConfigError(GeniePyError):
    """Configuration error."""



"""Module to test online sources parsers."""
import pytest
from geniepy.datamgmt.scrapers import PubMedScraper
from geniepy.datamgmt.parsers import BaseParser, PubMedParser
from geniepy.errors import ParserError
import tests.testdata as td


VALID_DF = td.PUBMED_VALID_DF
INVALID_DF = td.PUBMED_INVALID_DF


class TestPubMedParser:
    """Pytest PubMed Parser class."""

    parser: BaseParser = PubMedParser()
    scraper: PubMedScraper = PubMedScraper()
    parser.scraper = scraper

    def test_constructor(self):
        """Ensure scraper obj constructed successfully."""
        assert self.parser is not None

    @pytest.mark.parametrize("payload", INVALID_DF)
    def test_invalid_payload(self, payload):
        """Test invalid dataframe."""
        # Should return list with errors
        assert self.parser.validate(payload)

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_valid_payload(self, payload):
        """Test valid dataframe."""
        # Should return empty list
        assert not self.parser.validate(payload)

    @pytest.mark.slow_integration_test
    @pytest.mark.parametrize("chunksize", [1, 10, 100])
    def test_parse_valid(self, chunksize):
        """Test parsing valid recrods."""
        scrape_gen = self.parser.scraper.scrape(chunksize=chunksize)
        xml_articles = next(scrape_gen)
        parsed_df = self.parser.parse(xml_articles)
        assert parsed_df.shape[0] == chunksize



"""Implementation of predictive classifiers."""
from abc import ABC, abstractmethod
import pandas as pd
from google_drive_downloader import GoogleDriveDownloader as gdd
from joblib import load as jload
from geniepy.errors import ClassifierError
import geniepy.config as gc

ERROR_SCORE = float(-1)

PCPCLSFR_NAME = "pub_score"
CTCLSFR_NAME = "ct_score"


class BaseClassifier(ABC):
    """Base Classifier Abstract Class."""

    __slots__ = [
        "_is_trained",  # True if classifier is trained
        "_model",  # The classifier model
        "_name",  # The name of the classifier (used as name of prediction col)
    ]

    def __init__(self, name):
        """
        Construct classifier obj.

        Arguments:
            name {[type]} -- The classifier's name
        """
        self._name = name
        self._is_trained = False
        self._model = None

    @property
    def name(self):
        """Return name prediction column."""
        return self._name

    @property
    def is_trained(self):
        """Return true is model is trained."""
        return self._is_trained

    @abstractmethod
    def predict(self, features: pd.DataFrame) -> float:
        """
        Calculate publication count label.

        It is import for this method to always produce results and handle exceptions
        internally, so execution isn't halted due to exceptions.

        Arguments:
            features {pd.DataFrame} -- Pandas series necessary prediction data

        Returns:
            float -- the classifier prediction
        """

    def load(self):
        """
        Load classifier model.

        Raises:
            ClassifierError -- If model doesn't load successfully
        """
        # Download model from google drive
        model_path = gc.TMP_DIR.joinpath("gene_disease_gbc.joblib")
        model_id = gc.get_model()
        try:
            gdd.download_file_from_google_drive(
                file_id=model_id, dest_path=model_path, unzip=True,
            )
            self._model = jload(model_path)
            self._is_trained = True
            model_path.unlink()
        except Exception:
            # If load fail
            raise ClassifierError("Unable to load model")


class Classifier(BaseClassifier):
    """Implementation of Publication Count Predictive Classifier."""

    def predict(self, features: pd.DataFrame):
        """
        Calculate publication count label.

        It is import for this method to always produce results and handle exceptions
        internally, so execution isn't halted due to exceptions.

        Arguments:
            features {pd.DataFrame} -- Pandas series necessary prediction data

        Returns:
            float -- the classifier prediction
        """
        if (not self._is_trained) or (features is None):
            # TODO log event "Received None features to predict"
            # TODO log event "Untrained classifier can't calculate predictions"
            return ERROR_SCORE
        # Only keep expected columns
        features.dropna(inplace=True)
        scores_df = pd.DataFrame(features["gene_disease_relationship"])
        filtered_features = features[
            [
                "num_publications",
                "citations_cum_sum",
                "authors_cum_sum",
                "chemicals_cum_sum",
                "cum_sum_journals",
                "num_languages",
                "sjr",
                "h_index",
                "us_published",
                "us_uk_published",
            ]
        ]
        # Predictions and probabilities
        predictions = self._model.predict(filtered_features)
        probs = self._model.predict_proba(filtered_features)
        prob_1 = [item[0] for item in probs]  # Get positiive probs only
        # Add new fields into original df
        scores_df["classifier_prediction"] = predictions
        scores_df["classifier_prob"] = prob_1
        return scores_df



"""
GeniePy Configuration Module.

This module is intended to provide functions that interprets content from the
package configuration file and generate the corresponding python objects.
"""
from shutil import copyfile
import yaml
from pathlib import Path
from geniepy.errors import ConfigError
import logging
import sys
from logging.handlers import TimedRotatingFileHandler

CONFIG_DIR = Path("~/.geniepy.d/").expanduser().resolve()
CONFIG_DIR.mkdir(parents=True, exist_ok=True)
TMP_DIR = CONFIG_DIR.joinpath("tmp").resolve()
TMP_DIR.mkdir(parents=True, exist_ok=True)

CONFIG_NAME = "config.yaml"
CONFIG_FILE = CONFIG_DIR.joinpath(CONFIG_NAME).resolve()
DEFAULT_CONFIG = Path(__file__).parent.joinpath(CONFIG_NAME).resolve()

LOG_FORMATTER = logging.Formatter(
    "%(asctime)s — %(name)s — %(levelname)s — %(message)s"
)
LOG_NAME = "geniepy.log"
LOG_FILE = CONFIG_DIR.joinpath(LOG_NAME)
LOG_FILE.touch()

# Check for config.yaml in geniepy dir. Otherwise, create default
if not CONFIG_FILE.exists():
    copyfile(DEFAULT_CONFIG, CONFIG_FILE)
    print(
        f"""Created default configuration. Please configure geniepy and
add your Google Big Query credentials. {str(CONFIG_FILE)}"""
    )


def read_yaml() -> dict:
    """Read yaml configuration file and return dict."""
    with open(CONFIG_FILE) as config_file:
        return yaml.load(config_file, Loader=yaml.FullLoader)


def get_model() -> str:
    """Return the model file_id."""
    configdict = read_yaml()
    return configdict["model"]


def get_projname() -> str:
    """Retrieve gbq project name from config file."""
    configdict = read_yaml()
    return configdict["gbq"]["proj"]


def get_dataset(dsname: str) -> str:
    """Retrive gbq dataset from config file. 'master' or 'scoring'."""
    configdict = read_yaml()
    return configdict["gbq"]["dataset"][dsname]


def get_chunksize() -> int:
    """Retrieve standard genie generators chunk size."""
    configdict = read_yaml()
    return configdict["chunksize"]


def get_max_workers() -> int:
    """Retrieve standard genie generators maximum parallel processes."""
    configdict = read_yaml()
    return configdict["max_workers"]


def get_credentials() -> str:
    """Get credentials file path from config."""
    configdict = read_yaml()
    credentials_file = Path(configdict["gbq"]["credentials"]).expanduser()
    credentials_path = Path.cwd().joinpath(credentials_file).resolve()
    if not credentials_path.exists():
        # TODO log error properly
        print(f"Invalid configuration file: {credentials_path}")
        raise ConfigError("Credentials path not found")
    return credentials_path


def get_pubmed_ftp_server() -> str:
    """Retrieve PubMed FTP Server."""
    configdict = read_yaml()
    return configdict["pubmed_ftp_server"]


def get_pubmed_baseline_dir() -> str:
    """Retrieve PubMed FTP Directory for Baseline Dataset."""
    configdict = read_yaml()
    return configdict["pubmed_baseline_dir"]


def get_pubmed_update_dir() -> str:
    """Retrieve PubMed FTP Directory for Update Dataset."""
    configdict = read_yaml()
    return configdict["pubmed_update_dir"]


def get_pubmed_data_file() -> str:
    """Retrieve path for PubMed data file."""
    configdict = read_yaml()
    return configdict["pubmed_data_file"]


def get_pubmed_download_dir() -> str:
    """Retrieve path where to download PubMed data files."""
    configdict = read_yaml()
    return configdict["pubmed_download_dir"]


def get_logger(logger_name: str):
    """Retrieve geniepy logger."""
    file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
    file_handler.setFormatter(LOG_FORMATTER)
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.propagate = False
    return logger



"""
Data Access Managers Module.

DAOs are responsible for scraping, parsing, storing and delivering a specific type of
data. i.e. Pubmed Publications, Clinical Trials, Gene-Disease Relationships.
"""
from typing import Generator
from abc import ABC
from pandas import DataFrame
from geniepy.errors import SchemaError
from geniepy.datamgmt.parsers import (
    BaseParser,
    CtdParser,
    PubMedParser,
    ClassifierParser,
    PubtatorGeneParser,
    PubtatorDiseaseParser,
    SjrParser,
)
import geniepy.datamgmt.repositories as dr


class BaseDao(ABC):
    """
    Data Access Object Abstract Base Class.

    Each DAO is composed of one or more repositories and a parser which is used to
    structure and validate the DAO's data. Consult the parser's schema to check the
    format of the data stored in the repositories.
    """

    _repository: dr.BaseRepository
    """Database repository used by DAO to store objects."""
    _parser: BaseParser
    """DAO's parser to scraping and validating data."""

    def __init__(self, repository: dr.BaseRepository):
        """Initialize DAO state."""
        self._repository = repository

    @property
    def query_all(self):
        """Generate query string to query entire table."""
        return self._repository.query_all

    def query_pkey(self, val) -> str:
        """
        Generate query by primary key string.

        Arguments:
            val  -- value of primary key

        Returns:
            str -- The query str
        """
        return self._repository.query_pkey(val)

    def download(self, chunksize: int, **kwargs):
        """
        Download new data from online sources if available.

        Keyword Arguments:
            chunksize {[type]} -- The download method can be very computationally and
            memory intentive since it could possibly need to download and parse all
            records if the tables are empty. The chunksize allows the caller to limit
            how much memory is processed at a time while downloading and parsing the
            data.
        """
        for chunk_df in self._parser.fetch(chunksize, **kwargs):
            self._repository.save(chunk_df)

    def purge(self):
        """Purge all dao's database records."""
        self._repository.delete_all()

    # pylint: disable=bad-continuation
    def query(self, query: str, chunksize: int) -> Generator[DataFrame, None, None]:
        """
        Query DAO repo and returns a generator of DataFrames with query results.

        Keyword Arguments:
            query {str} -- Query string.
            chunksize {int} -- Number of rows of dataframe per chunk

        Returns:
            Generator[DataFrame] -- Generator to iterate over DataFrame results.
        """
        # pylint: disable=no-member
        return self._repository.query(query=query, chunksize=chunksize)

    def save(self, payload: DataFrame):
        """
        Save payload to database given data is valid.

        Arguments:
            payload {DataFrame} -- payload to be saved to table

        Raises:
            SchemaError: dataframe does not conform to table schema.
        """
        # pylint: disable=no-member
        errors = self._parser.validate(payload)
        if errors:
            raise SchemaError
        self._repository.save(payload)

    @property
    def tablename(self):
        """Return the dao repo tablename."""
        # pylint: disable=no-member
        return self._repository.tablename


class CtdDao(BaseDao):
    """Implementation of CTD Data Access Object."""

    __slots__ = ["_repository"]

    _parser: CtdParser = CtdParser()


class PubMedDao(BaseDao):
    """Implementation of CTD Data Access Object."""

    __slots__ = ["_repository"]

    _parser: PubMedParser = PubMedParser()


class ClassifierDao(BaseDao):
    """Implementation of DAO to handle output data used by UI for visualizations."""

    __slots__ = ["_repository"]

    _parser: ClassifierParser = ClassifierParser()

    def download(self, chunksize):
        """Classifiers don't need scrapers, so method not implemented."""
        raise NotImplementedError


class PubtatorGeneDao(BaseDao):
    """Data Access Object for Pubtator PMID/GeneID records."""

    __slots__ = ["_repository"]

    _parser: PubtatorGeneParser = PubtatorGeneParser()


class PubtatorDiseaseDao(BaseDao):
    """Data Access Object for Pubtator PMID/DiseaseID records."""

    __slots__ = ["_repository"]

    _parser: PubtatorDiseaseParser = PubtatorDiseaseParser()


class SjrDao(BaseDao):
    """Data Access Object for Pubtator PMID/DiseaseID records."""

    __slots__ = ["_repository"]

    _parser: SjrParser = SjrParser()



import tests.resources.mock as mock
from geniepy.datamgmt.parsers import PubtatorGeneParser


def test_parse():
    """Test parsing pubtator."""
    parser = PubtatorGeneParser()
    scraper = mock.MockPubtatorGeneScraper()
    gen_df = scraper.scrape(mock.TEST_CHUNKSIZE)
    parsed_df = parser.parse(next(gen_df))
    assert parsed_df.shape[0] == 1



"""
Module to extract ISSN metadata for
all available PubMed articles.
"""

from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
import glob
import gzip
import os
import sys
import xml.etree.ElementTree as ET

# Constants
EXTENSION = "gz"
TAG_Article = "PubmedArticle"
TAG_PMID = "./MedlineCitation/PMID"
TAG_ISSN = "./MedlineCitation/Article/Journal/ISSN"
ATT_ISSN_TYPE = "IssnType"
_CSV_SEPRATOR = ","


def decompress_archive(file):
    with gzip.open(file, "rb") as f:
        return f.read()


def task(file):
    file_content = decompress_archive(file)
    xmlDoc = ET.fromstring(file_content)

    # get all articles
    articles = xmlDoc.findall(TAG_Article)
    if articles is None:
        return None

    # extract PMID & ISSN data
    data = []
    for article in articles:
        pmid = ""
        issn = ""
        issn_type = ""

        if not article.find(TAG_PMID) is None:
            pmid = article.find(TAG_PMID).text

        if not article.find(TAG_ISSN) is None:
            issn = article.find(TAG_ISSN).text.replace("-", "")
            issn_type = article.find(TAG_ISSN).get(ATT_ISSN_TYPE)

        if len(pmid) > 0 and len(issn) > 0:
            data.append([pmid, issn, issn_type])

    print(f"Processed data file: {os.path.basename(file)}, articles: {len(data)}")
    return data


def main(inDir, outDir, maxWorkers):
    # get list of pubmed data files
    file_path = os.path.join(inDir, f"*.{EXTENSION}")
    files = glob.glob(file_path)

    if not files or len(files) <= 0:
        return

    # start parallel processing
    start_time = datetime.now()
    with ProcessPoolExecutor(max_workers=maxWorkers) as executor:
        result = executor.map(task, files)

    # write results to csv file
    for value in result:
        out_file = f"out-issn-{datetime.now().microsecond}.csv"
        out_path = os.path.join(outDir, out_file)
        with open(out_path, "w") as file:
            file.write(
                "article_id"
                + _CSV_SEPRATOR
                + "issn_id"
                + _CSV_SEPRATOR
                + "issn_type"
                + "\n"
            )
            for article in value:
                file.write(
                    str(article[0])
                    + _CSV_SEPRATOR
                    + str(article[1])
                    + _CSV_SEPRATOR
                    + str(article[2])
                    + "\n"
                )

    tot_time = datetime.now() - start_time
    print(
        f"Processed {len(files)} PubMed data files in \
{round(tot_time.total_seconds()/60,2)} minutes"
    )


if __name__ == "__main__":
    _in_dir = ""
    _out_dir = ""
    _max_workers = ""

    ERROR_MSG = "Command line arguments expected: <path to data dir>, \
        <path to output dir>, <number of parallel processes (2-16)>"

    if not sys.argv or len(sys.argv) < 4:
        raise ValueError(ERROR_MSG)

    # check command line argument for input directory
    if os.path.isdir(sys.argv[1]):
        _in_dir = sys.argv[1]
    else:
        raise ValueError(f"Input data directory is not valid. {ERROR_MSG}")

    # check command line argument for output directory
    if os.path.isdir(sys.argv[2]):
        _out_dir = sys.argv[2]
    else:
        raise ValueError(f"Output data directory is not valid. {ERROR_MSG}")

    # check argument for number of parallel processes
    try:
        _max_workers = int(sys.argv[3])
        if _max_workers < 2 or _max_workers > 16:
            raise ValueError(f"Number of parallel processes is not valid. {ERROR_MSG}")
    except Exception:
        raise ValueError(f"Number of parallel processes is not valid. {ERROR_MSG}")

    main(_in_dir, _out_dir, _max_workers)



"""Module to test PubMed scraper."""
import pytest
from geniepy.datamgmt.scrapers import PubMedScraper


class TestPubMedScraper:
    """Pytest PubMed Scraper class."""

    scraper: PubMedScraper = PubMedScraper()

    def test_constructor(self):
        """Ensure scraper obj constructed successfully."""
        assert self.scraper is not None

    @pytest.mark.slow_integration_test
    @pytest.mark.parametrize("chunksize", [1, 10, 100])
    def test_scrape_update(self, chunksize):
        """Test scraping valid recrods -- for Daily Update dataset."""
        scrape_gen = self.scraper.scrape(chunksize=chunksize)
        articles = next(scrape_gen)
        assert len(articles) == chunksize

    @pytest.mark.slow_integration_test
    @pytest.mark.parametrize("chunksize", [1, 10, 100])
    def test_scrape_baseline(self, chunksize):
        """Test scraping valid recrods -- for Baselines dataset."""
        scrape_gen = self.scraper.scrape(chunksize=chunksize, baseline=True)
        articles = next(scrape_gen)
        assert len(articles) == chunksize



"""Test package configuration & yaml files."""
from pathlib import Path
import geniepy.config as config
from tests import get_resources_path

TEST_CONFIG_NAME = "testconfig.yaml"
TEST_CONFIG_PATH = Path(get_resources_path()).resolve().joinpath(TEST_CONFIG_NAME)

# Override default config path for tests
config.CONFIG_FILE = TEST_CONFIG_PATH


def test_chunksize():
    """Test retrieving chunksize."""
    expected = 10
    actual = config.get_chunksize()
    assert actual == expected
